"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.install = void 0;
const commander_1 = require("commander");
const index_1 = require("../../lib/index");
const lodash_1 = require("lodash");
const chalk_1 = require("chalk");
function install(program) {
    program
        .command("upload")
        .alias("up")
        .addArgument(new commander_1.Argument("[source]", "要上传的本地资源，支持单文件、文件夹、glob 表达式"))
        .addArgument(new commander_1.Argument("[target]", "保存到 COS 的路径，默认是根路径"))
        .addOption(new commander_1.Option("--ignore [ignore...]", "需要忽略的文件的 glob 表达式"))
        .addOption(new commander_1.Option("--rename [rename]", "是否对文件进行重命名，如果为是，默认使用 16 个小写字母和数字的随机组合，指定数字可以自定义长度").argParser((v) => parseInt(v, 10) || 16))
        .addOption(new commander_1.Option("--flat", "是否展开文件夹层级"))
        .addOption(new commander_1.Option("--show-progress", "是否以进度条的形式展示上传过程"))
        .addOption(new commander_1.Option("--cdn-purge-cache", "是否刷新 CDN 缓存"))
        .addOption(new commander_1.Option("--cdn-push-cache", "是否预热 CDN 缓存"))
        .addOption(new commander_1.Option("--dry-run", "只模拟上传过程，不实际上传"))
        .description("上传本地文件到腾讯云 COS")
        .action(async (source, target, opts) => {
        try {
            const { configFile, client: clientName } = program.opts();
            const { client, upload } = await (0, index_1.getCOSCMDConfig)(configFile);
            const clients = client
                ? ((0, lodash_1.isArray)(client) ? client : [client]).filter(({ enable, name }) => enable && (!clientName || name === clientName))
                : [];
            if (!clients.length) {
                throw new Error(`缺少 COS 客户端，请先在 "cos.config.js" 文件中新增客户端`);
            }
            let uploads = [];
            if (source) {
                uploads = [{ source, target, ...opts }];
            }
            else if (!upload) {
                throw new Error(`缺少 upload 配置，你需要设置要上传的本地文件`);
            }
            else {
                uploads = ((0, lodash_1.isArray)(upload) ? upload : [upload]).map((v) => ({ ...v, ...opts }));
            }
            await clients.reduce(async (clientPromise, clientConfig) => {
                await clientPromise;
                const client = new index_1.Client(clientConfig);
                await uploads.reduce(async (uploadPromise, uploadConfig) => {
                    await uploadPromise;
                    const uploadRes = await client.upload(uploadConfig);
                    const { files, startTime, endTime } = uploadRes;
                    const maxPathLength = Math.max(...files.map(({ rPath }) => rPath.length));
                    const totalCount = files.length;
                    const successCount = files.filter(({ cosError }) => !cosError).length;
                    const failCount = totalCount - successCount;
                    const duration = endTime - startTime;
                    const oneMinute = 60 * 1000;
                    if (uploadConfig.showProgress || uploadConfig.dryRun) {
                        console.log(files
                            .map(({ rPath, cosKey, url, cosError }) => `${rPath} ${(0, chalk_1.gray)(`${"-".repeat(maxPathLength - rPath.length + 3)}>`)} ${cosError ? (0, chalk_1.red)(cosKey) : chalk_1.green.underline(url)}`)
                            .join("\n"));
                    }
                    console.log((0, chalk_1.blue)(`${uploadConfig.dryRun ? "模拟" : ""}上传完成 耗时 ${duration < oneMinute
                        ? `${(duration / 1000).toFixed(2)} 秒`
                        : `${Math.floor(duration / oneMinute)} 分 ${Math.round((duration % oneMinute) / 1000)} 秒`} 文件总数 ${(0, chalk_1.cyan)(totalCount)} 成功 ${(0, chalk_1.green)(successCount)} 失败 ${(0, chalk_1.red)(failCount)}`));
                }, Promise.resolve());
            }, Promise.resolve());
        }
        catch (error) {
            console.log((0, chalk_1.red)("上传失败"), error);
        }
    });
}
exports.install = install;
//# sourceMappingURL=upload.js.map